"""Constants for the PurpleAir integration."""
import logging

DOMAIN = "purpleair"

LOGGER = logging.getLogger(__package__)

CONF_LAST_UPDATE_SENSOR_ADD = "last_update_sensor_add"
CONF_READ_KEY = "read_key"
CONF_SENSOR_INDICES = "sensor_indices"
