"""Constants for Google Sheets integration."""
from __future__ import annotations

DOMAIN = "google_sheets"

DEFAULT_NAME = "Google Sheets"
DEFAULT_ACCESS = "https://www.googleapis.com/auth/drive.file"
