"""Constants for the Skybell HD Doorbell."""
import logging
from typing import Final

DEFAULT_NAME = "SkyBell"
DOMAIN: Final = "skybell"

IMAGE_AVATAR = "avatar"
IMAGE_ACTIVITY = "activity"

LOGGER = logging.getLogger(__package__)
