"""Constants for the Slack integration."""
from typing import Final

ATTR_BLOCKS = "blocks"
ATTR_BLOCKS_TEMPLATE = "blocks_template"
ATTR_FILE = "file"
ATTR_PASSWORD = "password"
ATTR_PATH = "path"
ATTR_SNOOZE = "snooze_endtime"
ATTR_URL = "url"
ATTR_USERNAME = "username"
ATTR_USER_ID = "user_id"

CONF_DEFAULT_CHANNEL = "default_channel"

DATA_CLIENT = "client"
DEFAULT_NAME = "Slack"
DEFAULT_TIMEOUT_SECONDS = 15
DOMAIN: Final = "slack"

SLACK_DATA = "data"
DATA_HASS_CONFIG = "slack_hass_config"
