"""Constants used by the escea component."""

DOMAIN = "escea"
ESCEA_MANUFACTURER = "Escea"
ESCEA_FIREPLACE = "Escea Fireplace"
ICON = "mdi:fire"

DATA_DISCOVERY_SERVICE = "escea_discovery"

DISPATCH_CONTROLLER_DISCOVERED = "escea_controller_discovered"
DISPATCH_CONTROLLER_DISCONNECTED = "escea_controller_disconnected"
DISPATCH_CONTROLLER_RECONNECTED = "escea_controller_reconnected"
DISPATCH_CONTROLLER_UPDATE = "escea_controller_update"

TIMEOUT_DISCOVERY = 20
