"""Constants for Bravia TV integration."""
from __future__ import annotations

from typing import Final

ATTR_CID: Final = "cid"
ATTR_MAC: Final = "macAddr"
ATTR_MANUFACTURER: Final = "Sony"
ATTR_MODEL: Final = "model"

CONF_CLIENT_ID: Final = "client_id"
CONF_IGNORED_SOURCES: Final = "ignored_sources"
CONF_NICKNAME: Final = "nickname"
CONF_USE_PSK: Final = "use_psk"

DOMAIN: Final = "braviatv"
LEGACY_CLIENT_ID: Final = "HomeAssistant"
NICKNAME_PREFIX: Final = "Home Assistant"
