"""Constants for the LED BLE integration."""

from typing import Final

DOMAIN = "led_ble"

DEVICE_TIMEOUT = 30
LOCAL_NAMES = {"LEDnet", "BLE-LED", "LEDBLE", "Triones", "LEDBlue"}

UNSUPPORTED_SUB_MODEL = "LEDnetWF"

UPDATE_SECONDS = 15

DEFAULT_EFFECT_SPEED: Final = 50
