"""Constants for the QNAP QSW integration."""

from typing import Final

ATTR_MAX: Final = "max"
ATTR_MESSAGE: Final = "message"

DOMAIN: Final = "qnap_qsw"
MANUFACTURER: Final = "QNAP"

RPM: Final = "rpm"

QSW_COORD_DATA: Final = "coordinator-data"
QSW_COORD_FW: Final = "coordinator-firmware"
QSW_REBOOT = "reboot"
QSW_TIMEOUT_SEC: Final = 25
QSW_UPDATE: Final = "update"
