"""Constants for the Reolink Camera integration."""

DOMAIN = "reolink"
PLATFORMS = ["camera"]

CONF_USE_HTTPS = "use_https"
CONF_PROTOCOL = "protocol"

DEFAULT_PROTOCOL = "rtsp"
DEFAULT_TIMEOUT = 60

HOST = "host"
DEVICE_UPDATE_INTERVAL = 60
