"""Constants for the Mastodon integration."""

import logging
from typing import Final

LOGGER = logging.getLogger(__name__)

CONF_BASE_URL: Final = "base_url"
DEFAULT_URL: Final = "https://mastodon.social"
