"""Constants for the Moon integration."""
from typing import Final

from homeassistant.const import Platform

DOMAIN: Final = "moon"
PLATFORMS: Final = [Platform.SENSOR]

DEFAULT_NAME: Final = "Moon"
