"""Constants for the Nobø Ecohub integration."""

DOMAIN = "nobo_hub"

CONF_AUTO_DISCOVERED = "auto_discovered"
CONF_SERIAL = "serial"
CONF_OVERRIDE_TYPE = "override_type"
OVERRIDE_TYPE_CONSTANT = "Constant"
OVERRIDE_TYPE_NOW = "Now"

NOBO_MANUFACTURER = "Glen Dimplex Nordic AS"
ATTR_HARDWARE_VERSION = "hardware_version"
ATTR_SOFTWARE_VERSION = "software_version"
ATTR_SERIAL = "serial"
ATTR_TEMP_COMFORT_C = "temp_comfort_c"
ATTR_TEMP_ECO_C = "temp_eco_c"
ATTR_ZONE_ID = "zone_id"
